#include "THAtomic.h"
#include "THTensor.h"
#include "THVector.h"
#include "generic/simd/simd.h"

#include "THBlas.h"
#include "THLapack.h"
#include "THRandom.h"
#include "THTensorDimApply.h"
#include "THMath.h"

#include "generic/THTensor.c"
#include "THGenerateAllTypes.h"

#include "generic/THTensor.c"
#include "THGenerateHalfType.h"

#include "generic/THTensorCopy.c"
#include "THGenerateAllTypes.h"

#include "generic/THTensorCopy.c"
#include "THGenerateHalfType.h"

#include "generic/THTensorRandom.c"
#include "THGenerateAllTypes.h"

#include "generic/THTensorMath.c"
#include "THGenerateAllTypes.h"

#include "generic/THTensorConv.c"
#include "THGenerateAllTypes.h"

#include "generic/THTensorLapack.c"
#include "THGenerateFloatTypes.h"
